#!/usr/bin/perl
#	apt-cacher.pl - CGI to provide a local cache for debian packages and release files and .deb files
#
#  $Revision: 1.6 $
#  $Source: /xenu/nick/CVS-TREE/Src/Apt-cacher/apt-cacher.pl,v $
#  $Date: 2001/09/26 16:33:29 $
#
#	PATH_INFO=/www.domain.name/some/path/filename

=head1 NAME

 apt-cacher.pl - CGI to provide a cache for downloaded Debian packages

 Distributed under the terms of the GNU Public Licence (GPL).


=head1 SYNOPSIS

 copy apt-cacher.pl to your apache cgi-bin directory
 edit /etc/apt/sources.list
 apt-get update
 apt-get upgrade

=head1 DESCRIPTION

If you have two or more Debian GNU/Linux machines on a fast local
network and you wish to upgrade packages from the Internet, you
don't want to download every package several times.

apt-cacher.pl is a CGI which will keep a cache on disk of Debian Packages
and Releas files (including .debs) which have been received from Debian
distribution servers on the Internet. When an apt-get client issues
a request for a file to apt-cacher.pl, if the file is already on disk
it is served to the client immediately, otherwise it is fetched from
the Internet, saved on disk, and then served to the client. This means
that several Debian machines can be upgraded but each package need be
downloaded only once.

To use this CGI you need a web server which supports CGI and a local
directory with plenty of free space (100 - 200 Mbytes is usually
adequate).

=head1 INSTALLATION

Assuming your web server is called B<www.myserver.com:80>
and your cache directory is called B</home/me/cache>, then:

1. Copy apt-cacher.pl to your web server's cgi-bin directory

2. Make sure apt-cacher.pl is executable (chmod a+rx apt-cacher.pl)

3. Edit apt-cacher.pl and set $cache_path to /home/me/cache

4. Make sure apt-cacher.pl is ok to run (perl -Mstrict -wc apt-cacher.pl)

5. mkdir /home/me/cache /home/me/cache/tmp

6. Make sure your client machines can access http://www.myserver.com:80/cgi-bin/apt-cacher.pl

If the CGI is executed without arguments, it will return a text/plain
error message.

7. Edit your /etc/apt/sources.list files, as follows. Where a line says
something like:

deb http://http.us.debian.org/debian testing main contrib non-free

change this to:

deb http://www.myserver.com:80/cgi-bin/apt-cacher.pl/http.us.debian.org/debian testing main contrib non-free

8. Do "apt-get update" as root. This will prime the cache dorectory with the
Package or Package.gz and Release files from the servers you used to use
directly.

9. Do "ls -laR /home/me/cache" to verify that files have been received and
stored. The "/home/me/cache/tmp" directory should be empty after downloads
have completed.

10. Do "apt-get update; apt-get -u upgrade" to start upgrading each machine.

=head1 CACHE DIRECTORY CONTENTS

apt-cacher.pl considers all .deb files with exactly the same filename
should be the same package (for example vim-rt_5.8.007-4_all.deb) no
matter where they are downloaded from, so these files are stored in
the cache directory using just the filename.

Packages and Release files (including Packages.gz) are potentially
different for every server and directory, so these files are stored
in the cache directory with the full hostname and path to the file,
with all slashes B</> changed to underscores B<_> (in the same
manner as apt-get names the files in /var/lib/apt/lists).

=head1 BUGS and FEATURES

1. Only HTTP is supported at present (i.e. apt-cacher.pl cannot access an
FTP url)

2. apt-cacher.pl probably only works with the Apache webserver, because
it relies on the webserver supplying the PATH_INFO environment variable

3. apt-cacher.pl uses B<wget> to retrieve files, so wget must be
installed.

4. apt-cacher.pl currently retrieves the complete file before returning
any data to the client. This may cause the web server to timeout and
apt-get to return an error. If this happens, wait for the wget
transfer(s) to complete and then re-issue the same "apt-get" command.

5. When a timeout happens apt-cacher.pl sometimes does not move its
temporary file to the destination directory, which means that the
transfer(s) may complete but the results may not be cached.

6. apt-cacher.pl uses a lock file to ensure that at most one instance
is running at any time. This is needed to ensure that apt-cacher.pl
does not attempt to download the same file concurrently.

7. The cache directory is never cleaned out by this tool, and it does
not issue an HTTP GET-IF-MODIFIED-SINCE request for Packages{,.gz}
and Release files, which may be frequently modified, so the user should
delete these files from the cache occasionally:

  find $cache_dir/tmp -type f -ctime +1 -print -exec rm {} \;

  find $cache_dir -type f \( -name Package\* -o -name Release\* \) -ctime +7 -print -exec rm {} \;

8. apt-get can resume a partial failed transfer, however apt-cacher.pl
cannot.

=cut

my $path = $ENV{PATH_INFO};
my $cache_path = "/xenu/home/nick/tmp/apt-cache";

# use IO::Handle;
use Fcntl ':flock';
use POSIX;

# We have a problem for large packages. The problem is that it appears
# the apache webserver gets tired of waiting for us to return a document
# so it sends a SIGTERM and then soon afterward, a SIGKILL. This is
# apparently a configuration directive in mod_perl ?? Anyway, let's fork
# a child so apache's violence doesn't kill us...

if (fork() > 0) {
	# parent
	$SIG{'TERM'} = sub { print STDERR "--- apt-cacher.pl: Parent received SIGTERM, exiting\n"; sleep(4); exit(8); };
	wait();
	exit($?);
}

# child executes under here ...

my $unique_filename;
my $child_pid;
my $child_rc;

# $SIG{'PIPE'} = sub { open(E, ">>$cache_path/errs"); print E "$$ received SIGPIPE\n"; close(E); };
# $SIG{'PIPE'} = 'IGNORE';
my $sigpipe_received = 0;

$SIG{'PIPE'} = sub {
	print STDERR "--- apt-cacher.pl: received SIGPIPE\n";
	$sigpipe_received = 1;
};

sub term_handler {
	print STDERR "--- apt-cacher.pl: received SIGTERM, terminating\n";

	# Kill the wget process if running and unlink its output file

	kill('TERM', $child_pid) if ($child_pid);
	unlink($unique_filename) if ($unique_filename);
	exit(8);
};

$SIG{'TERM'} = \&term_handler;

$SIG{'QUIT'} = sub { print STDERR "--- apt-cacher.pl: received SIGQUIT\n"; };
$SIG{'INT'} = sub { print STDERR "--- apt-cacher.pl: received SIGINT\n"; };

$SIG{'CHLD'} = sub {
	print STDERR "--- apt-cacher.pl: received SIGCHLD\n";
	wait();
	$child_rc = $?;
	undef $child_pid;
};

print STDERR "\n--- apt-cacher.pl: called with $path\n";

die "apt-cacher.pl: No cache_path directory!\n" if (!-d $cache_path);
die "apt-cacher.pl: No cache_path/tmp directory!\n" if (!-d "$cache_path/tmp");

open(LOCK, ">$cache_path/lock") or die "apt-cacher.pl: Unable to open $cache_path/lock for write: $!\n";
if (!flock(LOCK, LOCK_EX)) {
	die "Unable to achieve lock on $cache_path/lock: $!\n";
}

print STDERR "--- apt-cacher.pl: Lock achieved\n";
# keep LOCK open so that at most one apt-cacher.pl can be running at any time

# Now parse the path
if ($path !~ m(^/.+/.+)) {
	print STDERR "--- apt-cacher.pl: usage error\n";

	print <<EOF;
Content-Type: text/plain
Expires: 0

Usage: GET .../apt-cacher.pl/www.domain.name/some/path/to/filename
EOF

	exit(0);
}

my($host,$uri) = ($path =~ m(^/([^/]+)(/.+)));

if ($host ne '' && $uri ne '') {
	my $cache_by_fn = 0;
	my ($filename) = ($uri =~ /\/([^\/]+)$/);
	my $content_type = "text/plain";
	my $content_encoding;
	my $new_filename = $filename;

	if ($filename =~ /\.gz$/) {
		$content_encoding = "gzip";
	} elsif ($filename =~ /\.deb$/) {
		$content_type = "application/x-debian-package";
		$cache_by_fn = 1;
	} else {
		$content_type = "text/plain";
	}

	if (!$cache_by_fn) {
		# Make a long filename so we can cache these files
		$new_filename = "$host$uri";
		$new_filename =~ s/\//_/g;
	}

	my $cached_file = "$cache_path/$new_filename";

	print STDERR "--- Looking for $cached_file\n";

	if (!-f $cached_file) {
		# Otherwise, try to pick it up ...

		my $url = "http://$host$uri";
		$unique_filename = "$cache_path/tmp/response.$$";

		print STDERR "--- Try to pick up $url\n";
		$child_pid = fork();
		barf("fork failed") if (!defined $child_pid);

		if ($child_pid == 0) {
			# This is the child

			# Search the PATH environment for the wget executable
			foreach my $dir (split(/:/, $ENV{PATH})) {
				my $fn = "$dir/wget";
				if (-f $fn && -x _) {
					# { } surround exec to prevent perl warning
					{
						exec($fn, '-c', '-s', '-O', $unique_filename, $url);
					};
				}
			}

			# dang, no wget in path
			barf("Unable to exec wget");
		}

		# Otherwise, we must be the parent

		# Child processes are reaped by signal handler
		while ($child_pid > 0) {
			pause();
		}

		my $rc = $child_rc;

		if ($rc == 0) {
			rename($unique_filename, $cached_file);
		} else {
			unlink($unique_filename);
		}

		print STDERR "--- Pick up $url as $unique_filename, return code $rc\n";
		if ($rc != 0) {
			# Output nothing and exit - this will be a 500 server error?
			exit(0);
		}

		print STDERR "--- Rename $unique_filename to $cached_file\n";
	}

	# If it's already been cached, output it immediately
	if (-f $cached_file) {
		print STDERR "--- $cached_file is cached, return it\n";
#		print "Content-Encoding: $content_encoding\n" if ($content_encoding);
#		print "Content-Type: $content_type\n\n";

		# Output it ...
		open(I, "<$cached_file") || barf("Unable to open $cached_file");
		my $http_response_line = <I>; # Ignore 1st line

		my $buf;
		my $n;
		while ($n = read(I, $buf, 4096) > 0 && !$sigpipe_received) {
			print $buf;
		}

		close(I);

		print STDERR "--- Exit (lock released)\n";

		exit(0);
	}

	# This part of the code should be unreachable 
	barf("unreachable: $cached_file");
}

exit(0);

sub barf {
	my $errs = shift;

	die "--- Argh: $errs\n";
}
